package be.cylab.java.wowa.training;

import be.cylab.java.roc.Roc;
import be.cylab.java.roc.RocCoordinates;
import com.owlike.genson.GenericType;
import com.owlike.genson.Genson;
import info.debatty.java.aggregation.WOWA;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.List;
import java.util.Random;

/**
 * Class Utils. Contain tools for Trainer algorithm
 */
public final class Utils {

    private Utils() {

    }

    /**
     * Normalize a weights vector.
     * @param weights
     * @return
     */
    public static double[] normalizeWeights(final double[] weights) {
        double sum_weight = Utils.sumArrayElements(weights);
        double[] weights_normalized = new double[weights.length];
        for (int i = 0; i < weights.length; i++) {
            weights_normalized[i] = weights[i] / sum_weight;
        }
        return weights_normalized;
    }

    /**
     * Find the max distance in a List of SolutionDistance.
     * @param solutions
     * @return
     */
    public static double findMaxDistance(
            final List<SolutionDistance> solutions) {
        double max = Double.NEGATIVE_INFINITY;
        for (SolutionDistance solution : solutions) {
            if (solution.getDistance() > max) {
                max = solution.getDistance();
            }
        }
        return max;
    }

    /**
     * Find the minimum distance in a List of SolutionDistance.
     * @param solutions
     * @return
     */
    public static double findMinDistance(
            final List<SolutionDistance> solutions) {
        double min = Double.POSITIVE_INFINITY;
        for (SolutionDistance solution : solutions) {
            if (solution.getDistance() < min) {
                min = solution.getDistance();
            }
        }
        return min;
    }

    /**
     * Compute the sum of all distance in a List of SolutionDistance.
     * @param solutions
     * @return
     */
    public static double sumTotalDistance(
            final List<SolutionDistance> solutions) {
        double sum = 0;
        for (SolutionDistance solution : solutions) {
            sum += solution.getDistance();
        }
        return sum;
    }

    /**
     * Compute the sum of all array's element.
     * @param array
     * @return
     */
    public static double sumArrayElements(final double[] array) {
        float sum = 0;
        for (double weight : array) {
            sum += weight;
        }
        return sum;
    }

    /**
     * Generate a random integer in a range.
     * @param min
     * @param max
     * @return
     */
    public static int randomInteger(final int min, final int max) {
        if (min >= max) {
            throw new IllegalArgumentException("Max must be greater then min");
        }
        Random rnd = new Random();
        return rnd.nextInt((max - min) + 1) + min;
    }

    /**
     * Import a json file.
     * @param filename
     * @return
     * @throws IOException if we cannot read file
     */
    public static List<double[]> convertJsonToDataForTrainer(
            final String filename)
            throws IOException {

        Genson genson = new Genson();
        String data_json = new String(Files.readAllBytes(Paths.get(filename)));
        List<double[]> data = genson.deserialize(
                data_json,
                new GenericType<List<double[]>>() { });
        return data;
    }

    /**
     * Import result file.
     * @param filename
     * @return
     * @throws IOException if we cannot read file
     */
    public static double[] convertJsonToExpectedForTrainer(
            final String filename) throws IOException {

        Genson genson = new Genson();
        String expected_json = new String(
                Files.readAllBytes(Paths.get(filename)));
        double[] expected = genson.deserialize(expected_json, double[].class);
        return expected;
    }

    /**
     * @param solution
     * @param data
     * @param expected
     * @return
     */
    public static  double computeAUC(
            final SolutionDistance solution,
            final List<double[]> data,
            final double[] expected) {
        if (data.size() != expected.length) {
            throw new IllegalStateException(
                    "Data and expected have different size");
        }
        boolean[] true_alert = convertExpectedToBooleanArrayTrueAlert(expected);
        double[] score = computeWOWAScoreWithData(solution, data);
        Roc roc = new Roc(score, true_alert);
        return roc.computeAUC();
    }

    /**
     * @param solution
     * @param data
     * @param expected
     * @return
     */
    public static List<RocCoordinates> computeRocPoints(
            final SolutionDistance solution,
            final List<double[]> data,
            final double[] expected,
            final boolean save_on_csv) {
        if (data.size() != expected.length) {
            throw new IllegalStateException(
                    "Data and expected have different size");
        }
        boolean[] true_alert = convertExpectedToBooleanArrayTrueAlert(expected);
        double[] score = computeWOWAScoreWithData(solution, data);
        Roc roc = new Roc(score, true_alert);
        List<RocCoordinates> coordinates = roc.computeRocPoints();
        if (save_on_csv) {
            DateFormat format = new SimpleDateFormat("yyyyMMdd_HH:mm:ss");
            Date date = new Date();
            be.cylab.java.roc.Utils.storeRocCoordinatesInCSVFile(coordinates,
                    "RocCoordinates_" + format.format(date) + ".csv");
        }
        return coordinates;
    }

    /**
     * @param solution
     * @param data
     * @return
     */
    private static double[] computeWOWAScoreWithData(
            final SolutionDistance solution,
            final List<double[]> data) {
        double[] score = new double[data.size()];
        WOWA wowa = new WOWA(solution.getWeightsW(), solution.getWeightsP());
        for (int i = 0; i < data.size(); i++) {
            score[i] = wowa.aggregate(data.get(i));
        }
        return score;
    }

    /**
     * @param expected
     * @return
     */
    private static boolean[] convertExpectedToBooleanArrayTrueAlert(
            final double[] expected) {
        boolean[] true_alert = new boolean[expected.length];
        for (int i = 0; i < expected.length; i++) {
            if (expected[i] == 1) {
                true_alert[i] = true;
            } else if (expected[i] == 0) {
                true_alert[i] = false;
            } else {
                throw new IllegalStateException(
                        "Expected must be equal to 0 or 1 to compute ROC");
            }
        }
        return true_alert;
    }

}
